from __future__ import annotations

import logging

import datasets
import tabulate

from st_visium_datasets.utils import sanitize_str, setup_logging
from st_visium_datasets.visium import visium

setup_logging(level="DEBUG")

logger = logging.getLogger(__name__)


def get_visium_dataset_path() -> str:
    """Load visium dataset path for usage with `datasets.load_dataset` from huggingface

    Example:
    >>> from st_visium_datasets import get_visium_dataset_path
    >>> import datasets
    >>> dataset = datasets.load_dataset(get_visium_dataset_path(), name="human-cerebellum")
    >>> builder = datasets.load_dataset_builder(get_visium_dataset_path(), name="human-cerebellum")
    """
    return visium.__file__


def load_visium_dataset_builder(name: str = "all", **kwargs):
    return datasets.load_dataset_builder(
        get_visium_dataset_path(),
        name=sanitize_str(name, sep="-"),
        **kwargs,
    )


def load_visium_dataset(name: str = "all", **kwargs):
    if split := kwargs.get("split"):
        if not split.startswith(visium.VisiumDatasetBuilder.DEFAULT_SPLIT):
            logger.warning(
                f"split '{split}' does not start with '{visium.VisiumDatasetBuilder.DEFAULT_SPLIT}'. "
                f"Only one split is provided: {visium.VisiumDatasetBuilder.DEFAULT_SPLIT}"
            )
            split = visium.VisiumDatasetBuilder.DEFAULT_SPLIT
    else:
        split = visium.VisiumDatasetBuilder.DEFAULT_SPLIT

    kwargs.update({"split": split})
    return datasets.load_dataset(
        get_visium_dataset_path(),
        name=sanitize_str(name, sep="-"),
        **kwargs,
    )


def list_visium_datasets() -> list[str]:
    return list(visium.VisiumDatasetBuilder.builder_configs.keys())


def gen_visium_dataset_stat(name: str = "all") -> dict[str, str | int | float]:
    """TODO: to complete with other interseting stats."""
    config: visium.VisiumDatasetBuilderConfig = (
        visium.VisiumDatasetBuilder.builder_configs[name]  # type: ignore
    )
    number_of_spots_under_tissue = sum(
        c.number_of_spots_under_tissue for c in config.visium_configs
    )
    number_of_genes_detected = sum(
        c.number_of_genes_detected for c in config.visium_configs
    )
    return {
        "name": name,
        "number_of_spots_under_tissue": number_of_spots_under_tissue,
        "number_of_genes_detected": number_of_genes_detected,
    }


def gen_visium_dataset_stat_table() -> str:
    names = list_visium_datasets()
    stats = [gen_visium_dataset_stat(name) for name in names]
    return tabulate.tabulate(stats, headers="keys", tablefmt="github")


if __name__ == "__main__":
    table = gen_visium_dataset_stat_table()
    print(table)



from __future__ import annotations

import tarfile
import typing as tp
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import fsspec
from rich.console import Group
from rich.live import Live
from rich.progress import (
    BarColumn,
    DownloadColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TransferSpeedColumn,
)

from st_visium_datasets.utils.data_file import DataFile
from st_visium_datasets.utils.utils import remove_suffix

DatasetsDict = tp.Dict[str, tp.Dict[str, DataFile]]
LocalDatatsetDict = tp.Dict[str, tp.Dict[str, Path]]


def _create_dl_progress(**kwargs) -> Progress:
    columns = (
        TextColumn("[bold blue][{task.fields[task_name]}][/bold blue]"),
        TextColumn("[bold green]{task.fields[dataset_name]}[/bold green]"),
        "•",
        TextColumn("{task.fields[name]}"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeElapsedColumn(),
    )
    kwargs.setdefault("expand", True)
    return Progress(*columns, **kwargs)


def _create_global_progress(**kwargs) -> Progress:
    columns = (
        TextColumn("[bold blue]Downloading datasets[/bold blue]"),
        BarColumn(bar_width=None),
        TaskProgressColumn(),
        "•",
        MofNCompleteColumn(),
        "files",
        "•",
        TimeElapsedColumn(),
    )
    kwargs.setdefault("expand", True)
    return Progress(*columns, **kwargs)


def _split_extensions(path: str | Path) -> tuple[str, str]:
    ext = "".join(Path(path).suffixes)
    return remove_suffix(str(path), ext), ext


def _download_file(
    dataset_name: str,
    name: str,
    file: DataFile,
    download_dir: Path,
    dl_progress: Progress,
    force_download: bool = False,
    buffer_size: int = 8192,
):
    _, ext = _split_extensions(file.url)
    save_path = download_dir / f"{file.md5sum}{ext}"
    if save_path.is_file() and not force_download:
        return save_path

    task_id = dl_progress.add_task(
        "",
        total=file.size,
        dataset_name=dataset_name,
        task_name="Download",
        name=name,
    )

    with fsspec.open(file.url, "rb") as src:
        with open(save_path, "wb") as fout:
            while chunk := src.read(buffer_size):  # type: ignore
                fout.write(chunk)
                dl_progress.update(task_id, advance=len(chunk))

    dl_progress.update(task_id, visible=False)
    return save_path


def _extract_file(
    dataset_name: str,
    name: str,
    filepath: Path,
    dl_progress: Progress,
    force_extract: bool = False,
    buffer_size: int = 8192,
) -> Path:
    if not filepath.is_file():
        raise FileNotFoundError(filepath)

    if not str(filepath).endswith((".tar.gz")):
        return filepath

    extract_dir = Path(remove_suffix(str(filepath), ".tar.gz"))
    if extract_dir.is_dir() and list(extract_dir.iterdir()) and not force_extract:
        return extract_dir

    extract_dir.mkdir(exist_ok=True, parents=True)

    task_id = dl_progress.add_task(
        "",
        dataset_name=dataset_name,
        task_name="Extract",
        name=name,
    )
    fileobj = dl_progress.open(filepath, "rb", task_id=task_id, buffering=buffer_size)
    with tarfile.open(fileobj=fileobj) as tar:
        tar.extractall(extract_dir)

    dl_progress.update(task_id, visible=False)
    return extract_dir


def _download_and_extract_file(
    dataset_name: str,
    name: str,
    file: DataFile,
    download_dir: Path,
    dl_progress: Progress,
    force_download: bool = False,
    force_extract: bool = False,
    buffer_size: int = 8192,
) -> Path:
    save_path = _download_file(
        dataset_name,
        name,
        file,
        download_dir,
        dl_progress,
        force_download=force_download,
        buffer_size=buffer_size,
    )
    return _extract_file(
        dataset_name,
        name,
        save_path,
        dl_progress,
        force_extract=force_extract,
        buffer_size=buffer_size,
    )


def download_visium_datasets(
    datasets: DatasetsDict,
    *,
    download_dir: str | Path,
    force_download: bool = False,
    force_extract: bool = False,
    disable_pbar: bool = False,
) -> LocalDatatsetDict:
    download_dir = Path(download_dir)
    download_dir.mkdir(exist_ok=True, parents=True)

    total_nb_files = sum(len(d) for d in datasets.values())
    global_progress = _create_global_progress(disable=disable_pbar)
    dl_progress = _create_dl_progress(disable=disable_pbar)

    progress_group = Group(global_progress, dl_progress)

    def _worker(dataset_name: str, name: str, file: DataFile):
        save_path = _download_and_extract_file(
            dataset_name,
            name,
            file,
            download_dir,
            dl_progress,
            force_download=force_download,
            force_extract=force_extract,
        )
        return (dataset_name, name, save_path)

    futures = []
    local_datasets = defaultdict(dict)
    task_id = global_progress.add_task("", total=total_nb_files)
    with Live(progress_group):
        with ThreadPoolExecutor() as executor:
            for dataset_name, files in datasets.items():
                for name, file in files.items():
                    assert isinstance(file, DataFile)
                    f = executor.submit(_worker, dataset_name, name, file)
                    futures.append(f)

            for future in as_completed(futures):
                dataset_name, name, save_path = future.result()
                local_datasets[dataset_name][name] = save_path
                global_progress.update(task_id, advance=1)

    return local_datasets



from __future__ import annotations

import logging
import re
from functools import lru_cache
from pathlib import Path

from datasets.utils.logging import get_verbosity
from rich.logging import RichHandler

import st_visium_datasets


def remove_prefix(s: str, prefix: str) -> str:
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def remove_suffix(s: str, suffix: str) -> str:
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def setup_logging(level: int | str | None = None) -> None:
    logging.basicConfig(
        level=level if level is not None else get_verbosity(),
        format="%(message)s",
        datefmt="[%x %X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )


def sanitize_str(s: str, sep: str = "_") -> str:
    # Replace camelCase with sep
    s = re.sub("(.)([A-Z][a-z]+)", r"\1{}\2".format(sep), s)
    s = re.sub("([a-z0-9])([A-Z])", r"\1{}\2".format(sep), s)
    s = s.lower()
    # Replace ' ', '.', '-', '_', '/', '(', ')', ''', ',' with sep
    s = re.sub(r"[ -._/,;:()\']", sep, s)
    # Remove leading _, -, sep
    s = s.strip("_").strip("-")
    # Deduplicate consecutive '_' and '-'
    s = re.sub(r"_{2,}", "_", s)
    s = re.sub(r"-{2,}", "-", s)
    return s


def get_nested_filepath(dirname: Path, filename: str) -> Path:
    """Return path to file in directory"""
    paths = list(dirname.glob(f"**/{filename}"))
    if len(paths) == 0:
        raise FileNotFoundError(f"no {filename} found in {dirname}")
    if len(paths) > 1:
        raise ValueError(f"multiple {filename} found in {dirname}: {paths}")
    return paths[0]


@lru_cache
def get_configs_dir() -> Path:
    return Path(st_visium_datasets.__file__).parent / "configs"



from __future__ import annotations

import json
import logging
import os
import typing as tp
from pathlib import Path

import datasets
import pandas as pd

from st_visium_datasets.base import (
    VisiumDatasetBuilderConfig,
    gen_builder_configs,
)
from st_visium_datasets.builder import build_spots_dataset
from st_visium_datasets.utils import download_visium_datasets

logger = logging.getLogger(__name__)


class VisiumDatasetBuilder(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = VisiumDatasetBuilderConfig
    BUILDER_CONFIGS = list(gen_builder_configs())
    DEFAULT_CONFIG_NAME = "all"
    DEFAULT_SPLIT = "default"

    def __init__(
        self,
        spot_diameter_px: int | tp.Literal["auto"] = "auto",
        pil_resize_longest: int | None = 3840,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._spot_diameter_px = spot_diameter_px
        self._pil_resize_longest = pil_resize_longest
        self._cache_datasets_dir = os.path.join(self._cache_dir_root, "visium_datasets")

    def _info(self):
        return datasets.DatasetInfo(
            description=(
                f"Visium datasets for spatial transcriptomics. This datasets is "
                f"a collection of the following datasets from 10x "
                f"Genomics: {self.builder_configs.keys()}"
            ),
            features=datasets.Features(
                {
                    "species": datasets.Value("string"),
                    "anatomical_entity": datasets.Value("string"),
                    "disease_state": datasets.Value("string"),
                    "spot_path": datasets.Value("string"),
                    "spot_bytes": datasets.Value(dtype="binary"),
                    "features_path": datasets.Value("string"),
                    "features": datasets.features.Sequence(
                        {
                            "feature_id": datasets.Value("string"),
                            "feature_type": datasets.Value(dtype="string"),
                            "gene": datasets.Value("string"),
                            "count": datasets.Value(dtype="int64"),
                        }
                    ),
                }
            ),
            license="Creative Commons",
        )

    def _split_generators(self, dl_manager: datasets.DownloadManager):
        self.config: VisiumDatasetBuilderConfig
        datasets_urls = {
            vc.name: {
                "tiff": vc.image_tiff,
                "feature_bc_matrix": vc.feature_barcode_matrix_filtered,
                "spatial": vc.spatial_imaging_data,
            }
            for vc in self.config.visium_configs
        }

        logger.info("Downloading Visium datasets...")
        dataset_paths: dict = download_visium_datasets(
            datasets_urls,
            download_dir=self._cache_downloaded_dir,
            force_download=dl_manager.download_config.force_download,
            force_extract=dl_manager.download_config.force_extract,
            disable_pbar=False,
        )

        dataset_dirs = []

        logger.info("Building 'spot' datasets...")
        for name, paths in dataset_paths.items():
            config = self.config.visium_configs[name]
            dataset_dir = build_spots_dataset(
                config,
                Path(self._cache_datasets_dir),
                spot_diameter_px=self._spot_diameter_px,  # type:ignore
                pil_resize_longest=self._pil_resize_longest,
                **paths,
            )
            dataset_dirs.append(dataset_dir)
        return [
            datasets.SplitGenerator(
                name=self.DEFAULT_SPLIT,
                gen_kwargs={"dataset_dirs": dataset_dirs},
            )
        ]

    def _generate_examples(self, dataset_dirs: list[Path]):
        for dataset_dir in dataset_dirs:
            config_path = dataset_dir / "config.json"
            spots_dir = dataset_dir / "spots"
            config = json.loads(config_path.read_text())
            species = config["species"] or "unknown"
            anatomical_entity = config["anatomical_entity"] or "unknown"
            disease_state = config["disease_state"] or "unknown"

            for spot_path in spots_dir.glob("*.npy"):
                features_path = spot_path.with_suffix(".csv")
                barcode = spot_path.stem
                data = {
                    "species": species,
                    "anatomical_entity": anatomical_entity,
                    "disease_state": disease_state,
                    "spot_path": str(spot_path),
                    "features_path": str(features_path),
                    "spot_bytes": spot_path.read_bytes(),
                    "features": pd.read_csv(features_path).to_dict("records"),
                }
                yield barcode, data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(config={self.config})"



from st_visium_datasets.utils.data_file import DataFile
from st_visium_datasets.utils.download import download_visium_datasets
from st_visium_datasets.utils.utils import (
    get_configs_dir,
    get_nested_filepath,
    sanitize_str,
    setup_logging,
)



from __future__ import annotations

import json
import math
from pathlib import Path

import pandas as pd

from st_visium_datasets.utils import get_nested_filepath


def load_scale_factors(spatial_dir: Path) -> dict[str, float]:
    filename = "scalefactors_json.json"
    scale_factors_path = get_nested_filepath(spatial_dir, filename)
    with open(scale_factors_path) as f:
        return json.load(f)


def get_spot_diameter_px(spatial_dir: Path) -> int:
    """Return spot diameter in pixels"""
    scale_factors = load_scale_factors(spatial_dir)
    return math.ceil(scale_factors["spot_diameter_fullres"])


def get_tissue_positions_df(spatial_dir: Path) -> pd.DataFrame:
    """Return spots.tsv.gz as dataframe"""
    filenames = ["tissue_positions_list.csv", "tissue_positions.csv"]
    kwargs_list = [
        {
            "names": [
                "barcode",
                "in_tissue",
                "array_row",
                "array_col",
                "pxl_row_in_fullres",
                "pxl_col_in_fullres",
            ],
        },
        {"header": 0},
    ]
    for filename, kwargs in zip(filenames, kwargs_list):
        try:
            filepath = get_nested_filepath(spatial_dir, filename)
            df: pd.DataFrame = pd.read_csv(filepath, **kwargs)
            break
        except FileNotFoundError:
            continue
    else:
        raise FileNotFoundError(f"no {filenames} found in {spatial_dir}")

    df = df.rename(columns={"pxl_row_in_fullres": "y", "pxl_col_in_fullres": "x"})
    df = df[df["in_tissue"] == 1].drop(columns=["in_tissue", "array_row", "array_col"])
    df = df.set_index("barcode")
    return df



from __future__ import annotations

import csv
import gzip
from pathlib import Path

import pandas as pd
import scipy.io
from scipy.sparse import coo_matrix

from st_visium_datasets.utils import get_nested_filepath


def load_feature_barcode_matrix_df(
    feature_bc_matrix_dir: Path,
) -> pd.DataFrame:
    """Feature-Barcode Matrices are a key output generated by 10x Genomics technology,
    particularly in single-cell RNA sequencing (scRNA-seq).
    They essentially represent a way to organize and quantify gene expression data
    from individual cells.

    The returned dataframe is in the following format:

    - Rows = features (genes): These refer to genes or genomic regions that are being
    measured. Each row in the dataframe represents a gene or genomic region.
    Each row contains the feature id, gene name and feature type.
    - Columns = barcodes (cells): These represent individual cells. Each column in
    the dataframe corresponds to a specific cell that was sequenced.
    - The values in the df typically represent the count of RNA molecules
    (gene expression) detected for each gene in each cell. So, each cell's column
    will have counts for the various genes in the rows.
    """
    if not feature_bc_matrix_dir.is_dir():
        raise ValueError(f"{feature_bc_matrix_dir} is not a directory")

    mat = _load_mat(feature_bc_matrix_dir)
    feature_ids, gene_names, feature_types = _load_features(feature_bc_matrix_dir)
    barcodes = _load_barcodes(feature_bc_matrix_dir)

    df: pd.DataFrame = pd.DataFrame.sparse.from_spmatrix(mat)
    df.columns = barcodes
    df.insert(loc=0, column="feature_id", value=feature_ids)
    df.insert(loc=1, column="feature_type", value=feature_types)
    df.insert(loc=2, column="gene", value=gene_names)
    df = df.set_index(["feature_id", "gene", "feature_type"])
    return df


def _load_mat(feature_bc_matrix_dir: Path) -> coo_matrix:
    """Read in MEX format matrix as table"""
    mat_path = get_nested_filepath(feature_bc_matrix_dir, "matrix.mtx.gz")
    return scipy.io.mmread(mat_path)


def _load_features(
    feature_bc_matrix_dir: Path,
) -> tuple[list[str], list[str], list[str]]:
    """Read in features.tsv.gz as table
    - list of transcript/feature ids, e.g. 'ENSG00000187634'
    - list of gene names, e.g. 'RER1'
    - list of feature types, e.g. 'Gene Expression'
    """
    features_path = get_nested_filepath(feature_bc_matrix_dir, "features.tsv.gz")
    with gzip.open(features_path, "rt") as f:
        reader = csv.reader(f, delimiter="\t")
        rows = list(reader)
    feature_ids, gene_names, feature_types = zip(*rows)
    return feature_ids, gene_names, feature_types


def _load_barcodes(feature_bc_matrix_dir: Path) -> list[str]:
    """Read in barcodes.tsv.gz as table

    list of barcodes, e.g. 'AAACATACAAAACG-1'
    """
    barcodes_path = get_nested_filepath(feature_bc_matrix_dir, "barcodes.tsv.gz")
    with gzip.open(barcodes_path, "rt") as f:
        reader = csv.reader(f, delimiter="\t")
        return sum(reader, [])


def load_prove_set_df(probe_set_filepath: Path) -> pd.DataFrame:
    """Probes that are predicted to have off-target activity to homologous genes
    or sequences are excluded from analysis by default (all probes are present in
    the raw matrix file). These probes are marked with FALSE in the included column
    of the probe set reference CSV. Any gene that has at least one probe with predicted
    off-target activity will be excluded from filtered outputs. Setting filter-probes
    to false in the multi config file for cellranger multi will result in UMI counts
    from all non-deprecated probes, including those with predicted off-target activity,
    to be used in the analysis. Probes whose ID is prefixed with DEPRECATED are always
    excluded from the analysis. Please see the Probe Set Overview for details on the
    probes."""
    df = pd.read_csv(probe_set_filepath, comment="#")
    df = df.rename(columns={"gene_id": "feature_id"})
    df = df[df["included"]].drop(columns=["included"])
    return df



from pydantic import BaseModel, ConfigDict, Field


class DataFile(BaseModel, frozen=True):
    url: str
    md5sum: str
    size: int = Field(..., alias="bytes")

    model_config = ConfigDict(populate_by_name=True)

    def __repr__(self) -> str:
        return f"DataFile('{self.url}')"

    def __str__(self) -> str:
        return str(self.url)



from st_visium_datasets.load import (
    get_visium_dataset_path,
    list_visium_datasets,
    load_visium_dataset,
    load_visium_dataset_builder,
)
from st_visium_datasets.utils import setup_logging



from __future__ import annotations

import typing as tp
from pathlib import Path

import numpy as np
import tifffile
from PIL import Image, ImageDraw

from st_visium_datasets.base import VisiumConfig
from st_visium_datasets.feature_barcode import load_feature_barcode_matrix_df
from st_visium_datasets.spatial import get_spot_diameter_px, get_tissue_positions_df


def build_spots_dataset(
    config: VisiumConfig,
    data_dir: Path,
    *,
    tiff: Path,
    feature_bc_matrix: Path,
    spatial: Path,
    spot_diameter_px: int | tp.Literal["auto"] = "auto",
    pil_resize_longest: int | None = 3840,
) -> Path:
    dataset_dir = data_dir / config.name
    spots_dir = dataset_dir / "spots"
    spots_dir.mkdir(exist_ok=True, parents=True)

    img: np.ndarray = _read_tiff(tiff)

    features_df = load_feature_barcode_matrix_df(feature_bc_matrix)
    if spot_diameter_px == "auto":
        spot_diameter_px = get_spot_diameter_px(spatial)
    spots_df = get_tissue_positions_df(spatial)

    pil_img, resize_ratio = _get_pil_img(img, resize_longest=pil_resize_longest)
    draw, color = ImageDraw.Draw(pil_img), "blue"
    # line width of 2px and point radius of 3px if max size is 3840
    line_width, point_radius = (
        (max(pil_img.size) * 2) // 3840,
        (max(pil_img.size) * 3) // 3840,
    )

    for barcode, row in spots_df.iterrows():
        x, y = row["x"], row["y"]
        xmin, ymin, xmax, ymax = _get_spot_bbox(x, y, spot_diameter_px)
        spot_img: np.ndarray = img[ymin:ymax, xmin:xmax, ...]
        spot_img_path = spots_dir / f"{barcode}.npy"
        np.save(spot_img_path, spot_img)

        # get spot features
        features_df[[barcode]].rename(columns={barcode: "count"}).to_csv(
            spots_dir / f"{barcode}.csv"
        )

        _draw_spot_bbox(draw, (xmin, ymin, xmax, ymax), resize_ratio, color, line_width)
        _draw_spot_center(draw, (x, y), resize_ratio, color, point_radius)

    pil_img.save(dataset_dir / "spots.png", format="PNG")
    config.save(dataset_dir / "config.json")
    return dataset_dir


def _get_spot_bbox(x: int, y: int, spot_diameter_px: int) -> tuple[int, int, int, int]:
    spot_radius = spot_diameter_px / 2
    xmin, ymin = int(x - spot_radius), int(y - spot_radius)
    xmax, ymax = int(x + spot_radius), int(y + spot_radius)
    return xmin, ymin, xmax, ymax


def _read_tiff(tiff: Path) -> np.ndarray:
    return tifffile.imread(tiff, is_ome=False)


def _get_pil_img(
    img: np.ndarray, resize_longest: int | None = 3840
) -> tuple[Image.Image, float]:
    pil_img = Image.fromarray(img, mode="YCbCr").convert("RGB")
    if not resize_longest:
        return pil_img, 1.0
    w, h = pil_img.size
    resize_ratio = resize_longest / max(h, w)
    new_w, new_h = int(w * resize_ratio), int(h * resize_ratio)
    return pil_img.resize((new_w, new_h), resample=Image.BILINEAR), resize_ratio


def _draw_spot_bbox(
    draw: ImageDraw.ImageDraw,
    bbox: tuple[int, int, int, int] | np.ndarray,
    resize_ratio: float = 1.0,
    color: str = "blue",
    line_width=2,
) -> None:
    bbox = np.array(bbox) * resize_ratio
    draw.rectangle(tuple(bbox), outline=color, width=line_width)  # type: ignore


def _draw_spot_center(
    draw: ImageDraw.ImageDraw,
    center: tuple[int, int] | np.ndarray,
    resize_ratio: float = 1.0,
    color: str = "blue",
    point_radius=3,
) -> None:
    center = np.array(center) * resize_ratio
    draw.ellipse(
        (
            center[0] - point_radius,
            center[1] - point_radius,
            center[0] + point_radius,
            center[1] + point_radius,
        ),
        fill=color,
        outline=color,
    )



import typing as tp
from pathlib import Path

import datasets
import typing_extensions as tx
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, RootModel

from st_visium_datasets.utils import DataFile, get_configs_dir, sanitize_str
from st_visium_datasets.utils.utils import remove_prefix


class VisiumConfig(BaseModel):
    name: str
    homepage: str
    visium_dataset_name: str
    title: str
    description: str
    species: str
    anatomical_entity: str
    number_of_spots_under_tissue: int
    number_of_genes_detected: int = Field(
        ...,
        validation_alias=AliasChoices("genes_detected", "total_genes_detected"),
        serialization_alias="number_of_genes_detected",
    )

    image_tiff: DataFile
    spatial_imaging_data: DataFile
    feature_barcode_matrix_filtered: DataFile

    model_config = ConfigDict(extra="allow", populate_by_name=True)

    @classmethod
    def load(cls, filepath: Path) -> tx.Self:
        return cls.model_validate_json((filepath).read_text())

    def save(self, filepath: Path) -> None:
        filepath.write_text(self.model_dump_json(indent=2))


class VisiumConfigs(RootModel):
    root: tp.List[VisiumConfig] = Field(..., min_length=1)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item: tp.Union[int, str]) -> VisiumConfig:
        if isinstance(item, int):
            return self.root[item]
        configs = [c for c in self.root if c.name == item]
        if not configs:
            raise KeyError(f"{item!r} not found")
        if len(configs) > 1:
            raise KeyError(f"multiple configs found for {item!r}")
        return configs[0]

    def __len__(self):
        return len(self.root)

    @classmethod
    def load(cls, dirpath: Path) -> tx.Self:
        configs = [VisiumConfig.load(path) for path in dirpath.glob("**/*.json")]
        return cls(root=configs)


class VisiumDatasetBuilderConfig(datasets.BuilderConfig):
    def __init__(self, visium_configs: VisiumConfigs, **kwargs):
        super().__init__(**kwargs)
        self.visium_configs = visium_configs

    @classmethod
    def load(cls, name: str, path: Path, **kwargs) -> tx.Self:
        if path.is_dir():
            visium_configs = VisiumConfigs.load(path)
        elif path.is_file():
            visium_configs = VisiumConfigs(root=[VisiumConfig.load(path)])
        else:
            raise FileNotFoundError(f"{path} is not a file or directory")

        name = sanitize_str(name, sep="-")
        return cls(name=name, visium_configs=visium_configs, **kwargs)

    def __repr__(self) -> str:
        return (
            f"VisiumDatasetBuilderConfig(name={self.name!r}, "
            f"{len(self.visium_configs)} visium sub-datasets)"
        )


def gen_builder_configs(
    configs_dir: tp.Optional[Path] = None,
) -> tp.Generator[VisiumDatasetBuilderConfig, None, None]:
    base_configs_dir = get_configs_dir()

    if configs_dir is None:
        yield VisiumDatasetBuilderConfig.load("all", base_configs_dir)
        configs_dir = configs_dir or base_configs_dir

    for path in configs_dir.iterdir():
        if path == base_configs_dir or path.name.startswith((".", "_")):
            continue

        if path.is_dir():
            name = remove_prefix(str(path), str(base_configs_dir))
            yield VisiumDatasetBuilderConfig.load(name, path)
            yield from gen_builder_configs(path)
